///////////////////////////////////////////////////////////////////////////////
//     dP                                                               dP
//     88                                                               88
//     88d888b. dP    dP 88d888b. .d8888b. 88d888b. 88d888b. .d8888b. d8888P
//     88'  `88 88    88 88'  `88 88ooood8 88'  `88 88'  `88 88ooood8   88
//     88    88 88.  .88 88.  .88 88.  ... 88       88    88 88.  ...   88
//     dP    dP `8888P88 88Y888P' `88888P' dP       dP    dP `88888P'   dP
//                   .88 88
//               d8888P  dP
//
// Welcome Citizen of the hypersociety ;)
// ...
// Created by Samuel Loup in Môtiers (Switzerland), the first april 2021
///////////////////////////////////////////////////////////////////////////////

//add some electricity:
const watt = require("watt");

//seeder
const { v4: uuid } = require("uuid");

//usage
const yargs = require("yargs");
const argv = yargs
  .option("realm", {
    alias: "r",
    description: "realm to join",
    type: "string",
  })
  .option("society", {
    alias: "s",
    description: "society name",
    type: "string",
  })
  .option("port", {
    alias: "p",
    description: "listen",
    type: "integer",
  })
  .help()
  .default({ port: 2300 })
  .demandOption(["s"])
  .alias("help", "h").argv;

const { realm, society, port } = argv;
//open the matrix
const telnet = require("telnet2");
const blessed = require("blessed");

class Hypernet {
  constructor() {
    if (Hypernet._instance) {
      throw new Error("hypernet can't be instantiated more than once.");
    }
    Hypernet._instance = this;
    this.clients = new Map();
    process.on("SIGINT", function () {
      console.log("Caught interrupt signal");
      process.exit(0);
    });

    watt.wrapAll(this);
  }

  *bootStorage() {
    //hyperstuff
    const { Repo } = require("hypermerge");
    const Hyperswarm = require("hyperswarm");
    const os = require("os");
    const mkdirp = require("mkdirp");
    const path = require("path");
    const homePath = os.homedir();
    const storePath = path.join(homePath, "HyperSocietyData", society);
    yield mkdirp(storePath);
    console.log(`opening society repository: ${storePath}`);
    this.repo = new Repo({ path: storePath, port: 0 });
    this.swarm = Hyperswarm();
    this.repo.addSwarm(this.swarm);

    this.initialState = {
      orgs: {
        [society]: {
          id: society,
          createdAt: Date.now(),
        },
      },
      hypercards: {},
      realmEvents: {},
    };

    if (realm) {
      console.log(`joining realm: ${realm}`);
      this.realmKey = realm;
    } else {
      this.realmKey = this.repo.create(this.initialState);
      console.log(`opening realm: ${this.realmKey}`);
    }
    this.UI = require("./ui.js");
    this.repo.watch(this.realmKey, (state) => {
      console.log("watch");
      console.dir(state);
      this.ensureSociety(society, state);
      this.updateScreens(state);
    });

    this.repo.change(this.realmKey, (state) => {
      console.log("change");
      this.updateScreens(state);
    });
  }

  updateScreens(state) {
    for (const session of this.clients.values()) {
      const { id, screen } = session;
      console.log("updating screen:", id);
      screen.update(state);
      screen.render();
    }
  }

  ensureSociety(key, state) {
    if (!state.orgs[key]) {
      this.repo.change(this.realmKey, (state) => {
        state.orgs[key] = {
          id: key,
          createdAt: Date.now(),
        };
      });
    }
  }

  connect(client) {
    //term stuff
    const screen = blessed.screen({
      title: `Hyper Society :: v0.0.1`,
      smartCSR: true,
      input: client,
      output: client,
      terminal: "xterm-256color",
      fullUnicode: true,
    });
    const id = uuid();
    const sessionInfo = { id: id, connectedAt: Date.now(), screen };
    console.log("new session opening for ", id);

    this.repo.change(this.realmKey, (state) => {
      state.realmEvents[Date.now()] = {
        id,
        type: "connected",
      };
    });
    const { Box, Banner } = this.UI;
    const WelcomeBanner = (state) => `${Banner}

    ------------|{bold}net-infos{/bold}|---------
    ${this.realmKey}
    id: ${id}
    peers: ${this.swarm.peers}
    --------------------------------

    Bienvenue citoyen! 
    Tu es connecté sur l'HyperNet, a bientôt!

    Sociétés membres du royaumes:
    ${Object.keys(state.orgs)
      .map((s) => `\n    - ${s}`)
      .join("")}

    {bold}/Sam{bold}`;
    const welcomeScreen = Box.build(
      { content: WelcomeBanner(this.initialState), height: "100%" },
      screen
    );

    screen.update = (state) => {
      welcomeScreen.content = WelcomeBanner(state);
    };

    client.on("term", function (terminal) {
      screen.terminal = terminal;
      screen.render();
    });

    client.on("size", function (width, height) {
      client.columns = width;
      client.rows = height;
      client.emit("resize");
    });

    client.on("close", function () {
      if (!screen.destroyed) {
        screen.destroy();
      }
    });

    screen.key(["C-c", "q"], function (ch, key) {
      screen.destroy();
    });

    let disposed = false;
    screen.on("destroy", () => {
      if (!disposed) {
        disposed = true;
        this.repo.change(this.realmKey, (state) => {
          state.realmEvents[Date.now()] = {
            id,
            type: "disconnected",
          };
        });
        console.log("destroying session for ", id);
        this.clients.delete(id);
        if (client.writable) {
          client.destroy();
        }
      }
    });
    //register session
    this.clients.set(id, sessionInfo);
  }
}

const start = watt(function* () {
  console.log("starting hypernet service...");
  const cli = new Hypernet();
  yield cli.bootStorage();
  console.log("starting hypernet service...[DONE]");
  telnet({ tty: true }, (clientSession) => {
    console.log("incoming session!");
    console.log("connecting...");
    cli.connect(clientSession);
    console.log("connecting...[DONE]");
  }).listen(port);
  console.log(`listening on port: ${port}`);
});
start();
